#include "tile_item.h"
#include "tile_mesh.h"

// The vertex and fragment shaders.
static const char* const vertex_shader =
    //"varying vec4 n;"
    "attribute highp vec4 position;     \n"
    "attribute highp vec4 texcoord;     \n"
    "attribute highp vec4 normal;     \n"
    "uniform highp mat4 combined;      \n"
    "uniform highp vec4 scale;      \n"
    "void main() {                   \n"
    "  gl_Position = combined * (position * scale);\n"
    "}";
static const char* const fragment_shader =
    //"varying vec4 n;"
    "uniform lowp vec4 color; \n"
    "void main() {            \n"
    "  gl_FragColor = color;  \n"
    "}";

// Gets the size in bytes of an OpenGL type.
static int SizeOfType(GLenum type) {
  static int sizes[] = {
    sizeof(char),
    sizeof(unsigned char),
    sizeof(short),
    sizeof(unsigned short),
    sizeof(int),
    sizeof(unsigned int),
    sizeof(float),
    2,
    3,
    4,
    sizeof(double)
  };
  Q_ASSERT(type >= GL_BYTE && type <= 0x140A);  // Value of GL_DOUBLE
  return sizes[type - GL_BYTE];
}

static const QSGGeometry::AttributeSet& getTileMeshAttributes() {
  static QSGGeometry::Attribute data[] = {
    QSGGeometry::Attribute::create(0, tile_mesh.vertices_size, tile_mesh.vertices_gl_type, true),
    QSGGeometry::Attribute::create(1, tile_mesh.texcoords_size, tile_mesh.texcoords_gl_type),
    QSGGeometry::Attribute::create(2, tile_mesh.normals_size, tile_mesh.normals_gl_type),
  };
  static QSGGeometry::AttributeSet attributes = { tile_mesh.size, tile_mesh.stride, data };
  return attributes;
}

// --- QtQuick item ---

TileItem::TileItem(QQuickItem* parent)
    : QQuickItem(parent) {
  setFlag(ItemHasContents);
}

qreal TileItem::size() const {
  return size_;
}

void TileItem::setSize(qreal size) {
  if (size_ == size)
    return;

  size_ = size;
  update();
  emit sizeChanged();
}

// --- Scene graph geometry node ---

QSGNode* TileItem::updatePaintNode(QSGNode* oldNode, UpdatePaintNodeData* data) {
  TileNode* node = static_cast<TileNode*>(oldNode);
  if (!node)
    node = new TileNode(this);

  node->setSize(size_);
  node->update();
  return node;
}

TileNode::TileNode(TileItem* item)
    : QSGGeometryNode()
    , item_(item)
    , geometry_(getTileMeshAttributes(),    // Attributes
                tile_mesh.count,            // Number of vertices
                tile_mesh.count,            // Number of indices
                tile_mesh.indices_gl_type)  // OpenGL type of the indices
    , size_(0.0)
    , dirty_geometry_(true) {
  memcpy(geometry_.indexData(), tile_mesh.indices,
         tile_mesh.count * SizeOfType(tile_mesh.indices_gl_type));
  geometry_.setDrawingMode(GL_TRIANGLES);
  setGeometry(&geometry_);
  setMaterial(&material_);
  setFlag(UsePreprocess, false);
}

TileNode::~TileNode() {
}

void TileNode::setSize(float size) {
  if (size_ != size) {
    size_ = size;
    material_.setScale(size * 0.5f);
  }
}

void TileNode::preprocess() {
}

void TileNode::update() {
  if (dirty_geometry_) {
    memcpy(geometry_.vertexData(), tile_mesh.attributes, tile_mesh.count * tile_mesh.stride);
    dirty_geometry_ = false;
  }
}

// -- Scene graph material shader ---

QSGMaterialType TileShader::type;

void TileShader::updateState(
    const RenderState& state, QSGMaterial* newEffect, QSGMaterial* oldEffect) {
  TileMaterial* oldMaterial = static_cast<TileMaterial*>(oldEffect);
  TileMaterial* newMaterial = static_cast<TileMaterial*>(newEffect);

  const QColor& c = newMaterial->color();
  if (oldMaterial == 0 || c != oldMaterial->color() || state.isOpacityDirty()) {
    float opacity = state.opacity();
    QVector4D v(c.redF() * c.alphaF() * opacity, c.greenF() * c.alphaF() * opacity,
                c.blueF() * c.alphaF() * opacity, c.alphaF() * opacity);
    program()->setUniformValue(color_id_, v);
  }

  const float s = newMaterial->scale();
  if (oldMaterial == 0 || s != oldMaterial->scale()) {
    QVector4D scale(s, s, s, 1.0f);
    program()->setUniformValue(scale_id_, scale);
  }

  if (state.isMatrixDirty()) {
    program()->setUniformValue(combined_id_, state.combinedMatrix());
    program()->setUniformValue(modelview_id_, state.modelViewMatrix());
  }
}

char const* const* TileShader::attributeNames() const {
  static char const *const attributes[] = { "position", "texcoord", "normal", 0 };
  return attributes;
}

void TileShader::initialize() {
  combined_id_ = program()->uniformLocation("combined");
  modelview_id_ = program()->uniformLocation("modelview");
  color_id_ = program()->uniformLocation("color");
  scale_id_ = program()->uniformLocation("scale");
}

const char *TileShader::vertexShader() const {
  return vertex_shader;
}

const char* TileShader::fragmentShader() const {
  return fragment_shader;
}

// --- Scene graph material ---

TileMaterial::TileMaterial()
  : color_(QColor(127, 127, 255)) {
  // setFlag(RequiresFullMatrix);
}

void TileMaterial::setColor(const QColor& color) {
  color_ = color;
  setFlag(Blending, color_.alpha() != 0xff);
}

void TileMaterial::setScale(const float scale) {
  scale_ = scale;
}

QSGMaterialType* TileMaterial::type() const {
  return &TileShader::type;
}

QSGMaterialShader* TileMaterial::createShader() const {
  return new TileShader;
}

int TileMaterial::compare(const QSGMaterial* other) const {
  const TileMaterial *m = static_cast<const TileMaterial *>(other);
  if (scale_ - m->scale() == 0.0f)
    return color_.rgba() - m->color().rgba();
  else
    return 1;
}
